;;; BongoDB, an embedded document-based engine
;;; Copyright (C) 2016 by Javier Sancho Fernandez <jsf at jsancho dot org>
;;;
;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (bongodb)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 vlist)
  #:use-module (srfi srfi-9)
  #:use-module (srfi srfi-9 gnu)
  #:export (make-collection
	    collection?
	    count
	    insert
	    find))


;;; Collection Definition

(define-record-type collection
  (make-collection-record table)
  collection?
  (table get-table))

(define (make-collection)
  (make-collection-record vlist-null))

(define (count col)
  (vlist-length (get-table col)))

(set-record-type-printer! collection
  (lambda (record port)
    (format port
	    "#<collection with ~a documents>"
	    (count record))))


;;; Working with documents

(define (format-document document)
  "Return a vhash document ready to store in a collection"
  (document-with-id
   (cond ((vhash? document)
	  document)
	 (else
	  (alist->vhash document)))))

(define (document-with-id document)
  "Return always a document with an id"
  (or (and (vhash-assoc '_id document)
	   document)
      (vhash-cons '_id (gensym) document)))

(define (insert col . documents)
  "Insert documents into the collection and return the new collection"
  (cond ((null? documents)
	 (values col '()))
	(else
	 (let* ((document (format-document (car documents)))
		(docid (cdr (vhash-assoc '_id document)))
		(newcol (make-collection-record (vhash-cons docid document (get-table col)))))
	   (receive (rescol docids)
	       (apply insert (cons newcol (cdr documents)))
	     (values rescol (cons docid docids)))))))

(define (find col filter)
  "Query the collection and return the documents that match with filter"
  (let ((table (get-table col)))
    (vhash-fold
     (lambda (key document result)
       (cond ((match-document? document filter)
	      (cons (vhash->alist document) result))
	     (else
	      result)))
     '()
     table)))

(define (match-document? document filter)
  "Try to match the given document with a list of conditions"
  (cond ((null? filter)
	 #t)
	(else
	 (and
	  (equal? (vhash-assoc (caar filter) document) (car filter))
	  (match-document? document (cdr filter))))))


;;; Tools

(define (vhash->alist vhash)
  (vhash-fold-right
   (lambda (key value result) (assoc-set! result key value))
   '()
   vhash))
